#include <Arduino.h>
#include <SerialCommand.h>
#include <EEPROM.h>

#define BOARD_WITH_SENSOR_INPUTS

#define SW_VERSION "1.0"
#define SLAVE_ADDR "\x3e"
#define EEPROM_OFFSET 10
#define BUS_DELAY 5

/* Commmands */
#define CMD_INFO SLAVE_ADDR "I"  /* returns software version and build time */
#define CMD_READ SLAVE_ADDR "R"  /* return relay status and sensor status */
#define CMD_WRITE SLAVE_ADDR "W" /* write LED and relay status */
#define CMD_LOCK SLAVE_ADDR "L"  /* get/set channels lock */
#define CMD_SENSOR_EN SLAVE_ADDR "SE"
#define CMD_SENSOR_INV SLAVE_ADDR "SI"
#define CMD_SENSOR_DELAY SLAVE_ADDR "SD"
#define CMD_RESPONSE "r"

/* Pin configuration */
#define CLK_SHARED 13
#define DATA_OUT 11
#define LATCH595 10
#define DATA_IN 12
#define LOAD165 9
#define TXEN485 2
#define OUTREG_EN 3

/* Update intervals */
#define SREG_INTERVAL 100

uint8_t ledMask = 0,                                 /* Button LEDs status (output register) */
  buttonMask = 0,                                    /* Button status (input register) */
  sensorMask = 0,                                    /* Sensor status (input register) */
  relayMask = 0,                                     /* Relay status (output register) */
  lockMask = 0,                                      /* Channel lock */
  blinkMask = 0,                                     /* LED blink flag */
  sensorEnableMask = 0xFF,                           /* Enable sensor control */
  sensorInvertMask = 0;                              /* Invert sensor signal */
uint16_t sensorLock[8] = { 0, 0, 0, 0, 0, 0, 0, 0 }; /* temporary lock for sensor states */
uint8_t sensorMaskWhenLocked = 0;
uint16_t sensorLockPreset[8] = { 0, 10, 10, 10, 10, 10, 10, 10 };

long registersUpdate = -SREG_INTERVAL, ledBlink = 0, sensorLockPass = -1000;

SerialCommand cmd;

/***** Read parameters from EEPROM *****/
void loadSavedState() {
  uint8_t e_ledMask, e_relayMask, e_lockMask, e_sensorEnableMask, e_sensorInvertMask;

  /* last known LED state */
  e_ledMask = EEPROM.read(EEPROM_OFFSET);
  if ((e_ledMask + EEPROM.read(EEPROM_OFFSET + 1)) == 255)
    ledMask = e_ledMask;

  /* last known relay state */
  e_relayMask = EEPROM.read(EEPROM_OFFSET + 2);
  if ((e_relayMask + EEPROM.read(EEPROM_OFFSET + 3)) == 255)
    relayMask = e_relayMask;

  /* channel locks */
  e_lockMask = EEPROM.read(EEPROM_OFFSET + 4);
  if ((e_lockMask + EEPROM.read(EEPROM_OFFSET + 5)) == 255)
    lockMask = e_lockMask;

  /* sensor enabled */
  e_sensorEnableMask = EEPROM.read(EEPROM_OFFSET + 6);
  if ((e_sensorEnableMask + EEPROM.read(EEPROM_OFFSET + 7)) == 255)
    sensorEnableMask = e_sensorEnableMask;

  /* sensor signal inverted */
  e_sensorInvertMask = EEPROM.read(EEPROM_OFFSET + 8);
  if ((e_sensorInvertMask + EEPROM.read(EEPROM_OFFSET + 9)) == 255)
    sensorInvertMask = e_sensorInvertMask;

  /* sensor lock preset */
  for (uint8_t i = 0; i < 16; i += 2) {
    uint8_t lowBy, highBy;
    lowBy = EEPROM.read(i + EEPROM_OFFSET + 10);
    highBy = EEPROM.read(i + EEPROM_OFFSET + 11);

    if ((lowBy == 0xFF) && (highBy == 0xFF)) {
      lowBy = 10;  // default to 10 seconds
      highBy = 0;
    }

    sensorLockPreset[i / 2] = (highBy << 8) + lowBy;
  }
}

/***** Save LED status and relay status in EEPROM *****/
void saveState() {
  EEPROM.update(EEPROM_OFFSET, ledMask);
  EEPROM.update(EEPROM_OFFSET + 1, 255 - ledMask);
  EEPROM.update(EEPROM_OFFSET + 2, relayMask);
  EEPROM.update(EEPROM_OFFSET + 3, 255 - relayMask);
}

/***** Return device information *****/
void infoRequest() {
  delay(BUS_DELAY);
  digitalWrite(TXEN485, HIGH);

  Serial.print(CMD_INFO CMD_RESPONSE "\x20");
  Serial.print("SW Version: ");
  Serial.println(SW_VERSION);
  Serial.print(CMD_INFO CMD_RESPONSE "\x20");
  Serial.print("Build date: ");
  Serial.println(__DATE__ "\x20" __TIME__);

  Serial.flush();

  delay(BUS_DELAY);
  digitalWrite(TXEN485, LOW);
}

/***** Returns current channel output and sensor state over serial port *****/
void readRequest() {
  delay(BUS_DELAY);
  digitalWrite(TXEN485, HIGH);

  Serial.print(CMD_READ CMD_RESPONSE "\x20");

  /* Relay status */
  Serial.print(relayMask, HEX);
  Serial.print('\x20');
  Serial.print(255 - relayMask, HEX);
  Serial.print('\x20');

  /* Sensor status */
  Serial.print(sensorMask, HEX);
  Serial.print('\x20');
  Serial.print(255 - sensorMask, HEX);
  Serial.println();

  Serial.flush();
  
  delay(BUS_DELAY);
  digitalWrite(TXEN485, LOW);
}

/***** Set LED and relay over RS485 *****/
/* TODO: why set LED too? */
void writeRequest() {
  char *arg;
  uint8_t param[4] = { 0, 0, 0, 0 };

  /* expecting 4 arguments (each parameter twice) */
  for (uint8_t i = 0; i < 4; i++) {
    arg = cmd.next();

    /* no argument provided, return existing statuses */
    if (arg == NULL)
      break;

    param[i] = (uint8_t)strtol(arg, 0, 16);
  }

  if (param[0] + param[1] == 255) ledMask = param[0];
  if (param[2] + param[3] == 255) relayMask = param[2];

  saveState();

  delay(BUS_DELAY);
  digitalWrite(TXEN485, HIGH);
  Serial.print(CMD_WRITE CMD_RESPONSE "\x20");

  /* LED status */
  Serial.print(ledMask, HEX);
  Serial.print('\x20');
  Serial.print(255 - ledMask, HEX);
  Serial.print('\x20');

  /* Relay status */
  Serial.print(relayMask, HEX);
  Serial.print('\x20');
  Serial.print(255 - relayMask, HEX);
  Serial.println();

  Serial.flush();
  
  delay(BUS_DELAY);
  digitalWrite(TXEN485, LOW);
}

/* generic function to read/set single 8-bit parameter */
void singleParameterProcessRequest(uint8_t *p, uint8_t eepromOffset, char *cmdS) {
  char *arg;
  uint8_t param1, param2;

  arg = cmd.next();
  if (arg == NULL) param1 = 0;
  else {
    param1 = (uint8_t)strtol(arg, 0, 16);
    arg = cmd.next();
    if (arg == NULL) param2 = 0;
    else param2 = (uint8_t)strtol(arg, 0, 16);
  }

  if (param1 + param2 == 255) *p = param1;

  EEPROM.update(EEPROM_OFFSET + eepromOffset, *p);
  EEPROM.update(EEPROM_OFFSET + eepromOffset + 1, 255 - *p);

  delay(BUS_DELAY);
  digitalWrite(TXEN485, HIGH);

  Serial.print(cmdS);
  Serial.print('\x20');

  /* Lock flag */
  Serial.print(*p, HEX);
  Serial.print('\x20');
  Serial.print(255 - *p, HEX);
  Serial.println();

  Serial.flush();
  
  delay(BUS_DELAY);
  digitalWrite(TXEN485, LOW);
}

/***** Channel lock request *****/
void lockRequest() {
  singleParameterProcessRequest(&lockMask, 4, (char *)CMD_LOCK CMD_RESPONSE);
}

/***** Sensor enable flag *****/
void sensorEnableRequest() {
  singleParameterProcessRequest(&sensorEnableMask, 6, (char *)CMD_SENSOR_EN CMD_RESPONSE);
}

/***** Sensor signal invert flag *****/
void sensorInvertRequest() {
  singleParameterProcessRequest(&sensorInvertMask, 8, (char *)CMD_SENSOR_INV CMD_RESPONSE);
}

/***** Sensor delay *****/
void sensorDelayRequest() {
  char *arg;
  uint8_t param1;
  uint16_t param2;

  arg = cmd.next();
  if (arg != NULL) {
    param1 = atoi(arg);
    /* check if channel is valid */
    if ((param1 > 0) && (param1 <= 8)) {
      arg = cmd.next();

      if (arg != NULL) {
        param2 = (uint16_t)strtol(arg, 0, 16);
        if (param2 != 0xFFFF)
          sensorLockPreset[param1 - 1] = param2;
      }
    }
  }

  EEPROM.update(2 * (param1 - 1) + EEPROM_OFFSET + 10, uint8_t(param2 & 0xFF));
  EEPROM.update(2 * (param1 - 1) + EEPROM_OFFSET + 11, uint8_t(param2 >> 8));

  delay(BUS_DELAY);
  digitalWrite(TXEN485, HIGH);

  Serial.print(CMD_SENSOR_DELAY CMD_RESPONSE);
  Serial.print('\x20');

  for (uint8_t i = 0; i < 8; i++) {
    Serial.print(sensorLockPreset[i], HEX);

    if (i == 7) Serial.println();
    else Serial.print('\x20');
  }

  Serial.flush();
  
  delay(BUS_DELAY);
  digitalWrite(TXEN485, LOW);
}

/***** Blink button LEDs to signal error state *****/
void ledBlinker(uint8_t *leds, uint8_t *blink, uint8_t blinkCount = 10) {
  static uint8_t counter[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };

  for (uint8_t i = 0; i < 8; i++) {
    if (bitRead(*blink, i) == HIGH) {
      bitWrite(*leds, i, 1 - bitRead(*leds, i));
      counter[i]++;
    }

    if (counter[i] == blinkCount) {
      counter[i] = 0;
      bitClear(*blink, i);
    }
  }
}

void sensorLockTimePass() {
  for (uint8_t i = 0; i < 8; i++) {
    if (sensorLock[i] > 0)
      sensorLock[i] -= 1;
    else
      /* set the channel sensor state to initial value when sensor triggered an action */
      /* this is compared to actual sensor state, and will trigger an action since lock expired */
      bitWrite(sensorMask, i, bitRead(sensorMaskWhenLocked, i));
  }
}

void sensorMaskChanged(uint8_t newMask, uint8_t oldMask) {
  for (uint8_t i = 0; i < 8; i++) {
    /* check if current channel sensor state changed */
    if (bitRead(oldMask, i) != bitRead(newMask, i)) {
      /* is the sensor is enabled and there is no lock on current channel */
      if ((bitRead(sensorEnableMask, i) == HIGH) && (sensorLock[i] == 0)) {
        /* lock the channel */
        sensorLock[i] = sensorLockPreset[i];
        /* save current sensor states to compare with current state per channel when lock expires */
        sensorMaskWhenLocked = newMask;
        /* if sensor signal is inverted */
        if (bitRead(sensorInvertMask, i) == HIGH) {
          bitWrite(relayMask, i, 1 - bitRead(newMask, i));
          bitWrite(ledMask, i, 1 - bitRead(relayMask, i));
        } else {
          bitWrite(relayMask, i, bitRead(newMask, i));
          bitWrite(ledMask, i, bitRead(relayMask, i));
        }

        /* something changed; save to EEPROM */
        saveState();
      }
    }
  }
}

/***** Perform actions on button push *****/
void buttonMaskChanged(uint8_t newMask, uint8_t oldMask) {
  for (uint8_t i = 0; i < 8; i++) {
    /* check if button state changed from 0 to 1 (it has been pushed) */
    if ((bitRead(oldMask, i) == LOW) && (bitRead(newMask, i) == HIGH)) {
      /* check if there is any lock on the selected channel */
      if (bitRead(lockMask, i) == HIGH) {
        /* blink the LED */
        bitSet(blinkMask, i);
      } else {
        /* toggle relay state */
        bitWrite(relayMask, i, 1 - bitRead(relayMask, i));
        /* copy relay state to LED */
        bitWrite(ledMask, i, bitRead(relayMask, i));

        /* something changed; save to EEPROM */
        saveState();
      }
    }
  }
}

/***** Read/Write shift registers *****/
void readWriteShiftRegisters(uint8_t outLed, uint8_t outRelay, uint8_t *inButton, uint8_t *inSensor) {
  /* load sensor status in 74HC165 register */
  digitalWrite(LOAD165, LOW);
  digitalWrite(LOAD165, HIGH);

  /* Enable 74HC595 */
  digitalWrite(LATCH595, LOW);

  for (uint8_t i = 0; i < 8; i++) {
    bitWrite(*inSensor, i, 1 - digitalRead(DATA_IN));
    digitalWrite(DATA_OUT, bitRead(outLed, i));
    digitalWrite(CLK_SHARED, HIGH);
    digitalWrite(CLK_SHARED, LOW);
  }

#ifndef BOARD_WITH_SENSOR_INPUTS
  /* read again the same 74HC165 - for boards without sensor inputs */
  digitalWrite(LOAD165, LOW);
  digitalWrite(LOAD165, HIGH);
#endif

  for (uint8_t i = 0; i < 8; i++) {
    bitWrite(*inButton, i, 1 - digitalRead(DATA_IN));
    digitalWrite(DATA_OUT, 1 - bitRead(outRelay, i));
    digitalWrite(CLK_SHARED, HIGH);
    digitalWrite(CLK_SHARED, LOW);
  }

  /* Disable 74HC595 */
  digitalWrite(LATCH595, HIGH);
}

void setup() {
  pinMode(LATCH595, OUTPUT);
  pinMode(CLK_SHARED, OUTPUT);
  pinMode(DATA_OUT, OUTPUT);
  pinMode(LOAD165, OUTPUT);
  //pinMode(DATA_IN, INPUT_PULLUP);
  pinMode(DATA_IN, INPUT);

  digitalWrite(LATCH595, HIGH);
  digitalWrite(LOAD165, HIGH);

  pinMode(TXEN485, OUTPUT);
  digitalWrite(TXEN485, LOW);

  loadSavedState();
  readWriteShiftRegisters(ledMask, relayMask, NULL, &sensorMask);
  sensorMaskWhenLocked = sensorMask;


  Serial.begin(9600);

  cmd.addCommand((char *)CMD_INFO, infoRequest);
  cmd.addCommand((char *)CMD_READ, readRequest);
  cmd.addCommand((char *)CMD_WRITE, writeRequest);
  cmd.addCommand((char *)CMD_LOCK, lockRequest);
  cmd.addCommand((char *)CMD_SENSOR_EN, sensorEnableRequest);
  cmd.addCommand((char *)CMD_SENSOR_INV, sensorInvertRequest);
  cmd.addCommand((char *)CMD_SENSOR_DELAY, sensorDelayRequest);

  /* Enable output */
  pinMode(OUTREG_EN, OUTPUT);
  digitalWrite(OUTREG_EN, LOW);
}

void loop() {
  long tNow = millis();
  uint8_t _sensorMask, _buttonMask;

  cmd.readSerial();

  /* read and write shift registers; detect changes in inputs */
  if (tNow > registersUpdate + SREG_INTERVAL) {
    registersUpdate = tNow;
    readWriteShiftRegisters(ledMask, relayMask, &_buttonMask, &_sensorMask);

    if (_buttonMask != buttonMask) {
      buttonMaskChanged(_buttonMask, buttonMask);
      buttonMask = _buttonMask;
    }

    if (_sensorMask != sensorMask) {
      sensorMaskChanged(_sensorMask, sensorMask);
      sensorMask = _sensorMask;
    }
  }

  if (tNow > ledBlink + 250) {
    ledBlink = tNow;
    ledBlinker(&ledMask, &blinkMask);
  }

  if (tNow > sensorLockPass + 1000) {
    sensorLockPass = tNow;
    sensorLockTimePass();
  }
}